@@ -47,14 +47,6 @@ static const char unable_allocate_msg[] = "Unable to allocate output buffer.";
 #define KB (1024)
 #define MB (1024*1024)
 static const int BUFFER_BLOCK_SIZE[] =
-        /* If change this list, also change:
-             The CFFI implementation
-             OutputBufferTestCase unittest
-           If change the first blocks's size, also change:
-             ZstdDecompressReader.seek() method
-             ZstdFile.__init__() method
-             ZstdFile.read1() method
-             FileTestCase.test_decompress_limited() test */
         { 32*KB, 64*KB, 256*KB, 1*MB, 4*MB, 8*MB, 16*MB, 16*MB,
           32*MB, 32*MB, 32*MB, 32*MB, 64*MB, 64*MB, 128*MB, 128*MB,
           256*MB };
@@ -385,6 +377,9 @@ typedef struct {
     /* Ppmd8 context */
     CPpmd8 *cPpmd8;
 
+    /* End Mode */
+    Bool endmark;
+
     /* __init__ has been called, 0 or 1. */
     char inited;
     /* flush() has been called, 0 or 1. */
@@ -405,6 +400,9 @@ typedef struct {
     /* Ppmd7 context */
     CPpmd8 *cPpmd8;
 
+    /* End Mode */
+    Bool endmark;
+
     /* __init__ has been called, 0 or 1. */
     char inited;
     /* decode has been called with some data*/
@@ -461,8 +459,7 @@ Ppmd7Decoder_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     Ppmd7Decoder *self;
     self = (Ppmd7Decoder*)type->tp_alloc(type, 0);
     if (self == NULL) {
-        PyErr_NoMemory();
-        goto error;
+        return NULL;
     }
     assert(self->inited == 0);
     assert(self->inited2 == 0);
@@ -470,14 +467,13 @@ Ppmd7Decoder_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     /* Thread lock */
     self->lock = PyThread_allocate_lock();
     if (self->lock == NULL) {
-        PyErr_NoMemory();
         goto error;
     }
     return (PyObject*)self;
 
 error:
     Py_XDECREF(self);
-    return NULL;
+    return PyErr_NoMemory();
 }
 
 static void
@@ -776,7 +772,13 @@ Ppmd7Decoder_decode(Ppmd7Decoder *self,  PyObject *args, PyObject *kwargs) {
     reader.inBuffer = &in;
     self->rangeDec->Stream = (IByteIn *) &reader;
 
-    if (OutputBuffer_InitAndGrow(&buffer, &out, -1) < 0) {
+    int max_length;
+    if (length < 0) {
+        max_length = -1;
+    } else {
+        max_length = length;
+    }
+    if (OutputBuffer_InitAndGrow(&buffer, &out, max_length) < 0) {
         PyErr_SetString(PyExc_ValueError, "No Memory.");
         RELEASE_LOCK(self);
         return NULL;
@@ -794,6 +796,8 @@ Ppmd7Decoder_decode(Ppmd7Decoder *self,  PyObject *args, PyObject *kwargs) {
     }
     assert(self->inited2 == 1);
 
+    Bool result = True;
+    Py_BEGIN_ALLOW_THREADS
     if (data.len  > 0) {
         for (int i = 0; i < length; i++) {
             if (in.pos == in.size) {
@@ -801,13 +805,18 @@ Ppmd7Decoder_decode(Ppmd7Decoder *self,  PyObject *args, PyObject *kwargs) {
             }
             if (out.pos == out.size) {
                 if (OutputBuffer_Grow(&buffer, &out) < 0) {
-                    PyErr_SetString(PyExc_ValueError, "L616: Unknown status");
-                    goto error;
+                    PyErr_SetString(PyExc_ValueError, "No Memory.");
+                    result = False;
+                    break;
                 }
             }
             *((Byte *)out.dst + out.pos++) = Ppmd7_DecodeSymbol(self->cPpmd7, self->rangeDec);
         }
     }
+    Py_END_ALLOW_THREADS
+
+    if (!result)
+        goto error;
 
     ret = OutputBuffer_Finish(&buffer, &out);
 
@@ -909,16 +918,20 @@ static PyObject *
 Ppmd7Encoder_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     Ppmd7Encoder *self;
-    if ((self = (Ppmd7Encoder*)type->tp_alloc(type, 0)) != NULL) {
-        assert(self->inited == 0);
-        /* Thread lock */
-        if ((self->lock = PyThread_allocate_lock()) != NULL) {
-            return (PyObject*)self;
-        }
-        Py_XDECREF(self);
+    self = (Ppmd7Encoder*)type->tp_alloc(type, 0);
+    if (self == NULL) {
+        return NULL;
     }
-    PyErr_NoMemory();
-    return NULL;
+    assert(self->inited == 0);
+    /* Thread lock */
+    if ((self->lock = PyThread_allocate_lock()) == NULL) {
+        goto error;
+    }
+    return (PyObject*)self;
+
+error:
+    Py_XDECREF(self);
+    return PyErr_NoMemory();
 }
 
 static void
@@ -951,6 +964,7 @@ Ppmd7Encoder_init(Ppmd7Encoder *self, PyObject *args, PyObject *kwargs)
     static char *kwlist[] = {"max_order", "mem_size", NULL};
     PyObject *max_order = Py_None;
     PyObject *mem_size = Py_None;
+
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
                                      "OO:Ppmd7Encoder.__init__", kwlist,
                                      &max_order, &mem_size)) {
@@ -1001,6 +1015,7 @@ Ppmd7Encoder_init(Ppmd7Encoder *self, PyObject *args, PyObject *kwargs)
             }
         }
         PyMem_Free(self->cPpmd7);
+        PyErr_NoMemory();
     }
 
 error:
@@ -1039,17 +1054,23 @@ Ppmd7Encoder_encode(Ppmd7Encoder *self,  PyObject *args, PyObject *kwargs) {
     writer.outBuffer = &out;
     self->rangeEnc->Stream = (IByteOut *) &writer;
 
+    Bool result = True;
+    Py_BEGIN_ALLOW_THREADS
     for (UInt32 i = 0; i < data.len; i++){
         Ppmd7_EncodeSymbol(self->cPpmd7, self->rangeEnc, *((Byte *)data.buf + i));
         if (out.size == out.pos) {
             if (OutputBuffer_Grow(&buffer, &out) < 0) {
                 PyErr_SetString(PyExc_ValueError, "No memory.");
-                goto error;
+                result = False;
+                break;
             } else {
                 writer.outBuffer = &out;
             }
         }
     }
+    Py_END_ALLOW_THREADS
+    if (!result)
+        goto error;
 
     ret = OutputBuffer_Finish(&buffer, &out);
     RELEASE_LOCK(self);
@@ -1133,8 +1154,7 @@ Ppmd8Decoder_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     Ppmd8Decoder *self;
     self = (Ppmd8Decoder*)type->tp_alloc(type, 0);
     if (self == NULL) {
-        PyErr_NoMemory();
-        goto error;
+        return NULL;
     }
     assert(self->inited == 0);
     assert(self->inited2 == 0);
@@ -1142,14 +1162,13 @@ Ppmd8Decoder_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     /* Thread lock */
     self->lock = PyThread_allocate_lock();
     if (self->lock == NULL) {
-        PyErr_NoMemory();
         goto error;
     }
     return (PyObject*)self;
 
 error:
     Py_XDECREF(self);
-    return NULL;
+    return PyErr_NoMemory();
 }
 
 static void
@@ -1175,17 +1194,23 @@ PyDoc_STRVAR(Ppmd8Decoder_doc, "A PPMd compression algorithm decoder.\n\n"
                                  "mem_size:  max memory size in bytes the compressor is able to use, bigger values improve compression,\n"
                                  "           raging from 10kB to physical memory size.\n"
                                  "           Default size is 16MB.\n"
+                                 "end_mark:  Boolean when using endmark for compression.\n"
+                                 "           Default is False.\n"
+                                 "           When True,  encoder use \x01 as escape character and \x00 as endmark.\n"
+                                 "           When source data has a byte \x01, it is converted to \x01\x01 first then compressed."
                                  );
 
 static int
 Ppmd8Decoder_init(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs)
 {
-    static char *kwlist[] = {"max_order", "mem_size", NULL};
+    static char *kwlist[] = {"max_order", "mem_size", "end_mark", NULL};
     PyObject *max_order = Py_None;
     PyObject *mem_size = Py_None;
+    Bool endmark = False;
+
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
-                                     "OO:Ppmd8Decoder.__init__", kwlist,
-                                     &max_order, &mem_size)) {
+                                     "OO|p:Ppmd8Decoder.__init__", kwlist,
+                                     &max_order, &mem_size, &endmark)) {
         return -1;
     }
 
@@ -1196,6 +1221,8 @@ Ppmd8Decoder_init(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs)
     }
     self->inited = 1;
 
+    self->endmark = endmark;
+
     unsigned long maximum_order = 6;
     unsigned long memory_size = 16 << 20;
 
@@ -1230,6 +1257,7 @@ Ppmd8Decoder_init(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs)
             goto success;
         }
         PyMem_Free(self->cPpmd8);
+        PyErr_NoMemory();
     }
 
 error:
@@ -1246,7 +1274,7 @@ PyDoc_STRVAR(Ppmd8Decoder_flush_doc, "flush()\n"
 static PyObject *
 Ppmd8Decoder_flush(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs) {
     static char *kwlist[] = {"length", NULL};
-    int length;
+    int length = -1;
     PPMD_inBuffer in;
     BufferReader reader;
     BlocksOutputBuffer buffer;
@@ -1254,18 +1282,23 @@ Ppmd8Decoder_flush(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs) {
     PyObject *ret = NULL;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
-                                     "i:Ppmd8Decoder.flush", kwlist,
+                                     "|i:Ppmd8Decoder.flush", kwlist,
                                      &length)) {
         return NULL;
     }
 
     /* Only called once */
     if (self->flushed) {
-        PyErr_SetString(PyExc_RuntimeError, flush_twice_msg);
-        goto error;
+        return PyBytes_FromString("");
     }
     self->flushed = 1;
 
+    if (!self->endmark && length < 0) {
+       PyErr_SetString(PyExc_ValueError,
+                       "Length should be specified and a positive integer when no endmark mode.");
+       return NULL;
+    }
+
     if (self->inited2 == 0) {
        PyErr_SetString(PyExc_RuntimeError,
                        "Call flush() before calling decode()");
@@ -1305,15 +1338,41 @@ Ppmd8Decoder_flush(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs) {
         return NULL;
     }
 
+    Bool result = True;
+    Py_BEGIN_ALLOW_THREADS
     for (int i = 0; i < length; i++) {
         if (out.pos == out.size) {
             if (OutputBuffer_Grow(&buffer, &out) < 0) {
                 PyErr_SetString(PyExc_ValueError, "L1328: Unknown status");
-                goto error;
+                result = False;
+                break;
             }
         }
-        *((Byte *)out.dst + out.pos++) = Ppmd8_DecodeSymbol(self->cPpmd8);
+
+        if (self->endmark) {
+            unsigned char c = Ppmd8_DecodeSymbol(self->cPpmd8);
+            if (c != 0x01) {
+                *((Byte *)out.dst + out.pos++) = c;
+            } else {
+                c = Ppmd8_DecodeSymbol(self->cPpmd8);
+                if (c == 0x01) { // escaped character
+                    *((Byte *)out.dst + out.pos++) = c;
+                } else if (c == 0x00){ // endmark
+                    break;
+                } else {
+                    PyErr_SetString(PyExc_ValueError, "Corrupted input data.");
+                    result = False;
+                    break;
+                }
+            }
+        } else {
+            *((Byte *)out.dst + out.pos++) = Ppmd8_DecodeSymbol(self->cPpmd8);
+        }
     }
+    Py_END_ALLOW_THREADS
+    if (!result)
+        goto error;
+
     if (!self->cPpmd8->Code != 0) {
         PyErr_SetString(PyExc_ValueError, "Decompression failed.");
         goto error;
@@ -1341,7 +1400,7 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
     static char *kwlist[] = {"data", "length", NULL};
     BlocksOutputBuffer buffer;
     Py_buffer data;
-    int length;
+    int length = -1;
     PPMD_inBuffer in;
     PPMD_outBuffer out;
     PyObject *ret = NULL;
@@ -1349,7 +1408,7 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
     BufferReader reader;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
-                                     "y*i:Ppmd8Decoder.decode", kwlist,
+                                     "y*|i:Ppmd8Decoder.decode", kwlist,
                                      &data, &length)) {
         return NULL;
     }
@@ -1360,6 +1419,12 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
        return NULL;
     }
 
+    if (!self->endmark && length < 0) {
+       PyErr_SetString(PyExc_ValueError,
+                       "Length should be a positive integer when no endmark mode.");
+       return NULL;
+    }
+
     ACQUIRE_LOCK(self);
 
     /* Prepare input buffer w/wo unconsumed data */
@@ -1446,7 +1511,13 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
     reader.inBuffer = &in;
     self->cPpmd8->Stream.In = (IByteIn *) &reader;
 
-    if (OutputBuffer_InitAndGrow(&buffer, &out, -1) < 0) {
+    int max_length;
+    if (length < 0) {
+        max_length = -1;
+    } else {
+        max_length = length;
+    }
+    if (OutputBuffer_InitAndGrow(&buffer, &out, length) < 0) {
         PyErr_SetString(PyExc_ValueError, "No Memory.");
         RELEASE_LOCK(self);
         return NULL;
@@ -1464,20 +1535,46 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
     }
     assert(self->inited2 == 1);
 
+    Bool result = True;
+    Py_BEGIN_ALLOW_THREADS
     if (data.len  > 0) {
-        for (int i = 0; i < length; i++) {
+        int i = 0;
+        while (length == -1 || i++ < length ) {
             if (in.pos == in.size) {
                 break;
             }
             if (out.pos == out.size) {
                 if (OutputBuffer_Grow(&buffer, &out) < 0) {
                     PyErr_SetString(PyExc_ValueError, "L616: Unknown status");
-                    goto error;
+                    result = False;
+                    break;
                 }
             }
-            *((Byte *)out.dst + out.pos++) = Ppmd8_DecodeSymbol(self->cPpmd8);
+            if (self->endmark) {
+                unsigned char c = Ppmd8_DecodeSymbol(self->cPpmd8);
+                if (c != 0x01) {
+                    *((Byte *)out.dst + out.pos++) = c;
+                } else {
+                    c = Ppmd8_DecodeSymbol(self->cPpmd8);
+                    if (c == 0x01) { // escaped character
+                        *((Byte *)out.dst + out.pos++) = c;
+                    } else if (c == 0x00){ // endmark
+                        self->flushed = True;
+                        break;
+                    } else {
+                        PyErr_SetString(PyExc_ValueError, "Corrupted input data.");
+                        result = False;
+                        break;
+                    }
+                }
+            } else {
+                *((Byte *)out.dst + out.pos++) = Ppmd8_DecodeSymbol(self->cPpmd8);
+            }
         }
     }
+    Py_END_ALLOW_THREADS
+    if (!result)
+        goto error;
 
     ret = OutputBuffer_Finish(&buffer, &out);
 
@@ -1567,16 +1664,20 @@ static PyObject *
 Ppmd8Encoder_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     Ppmd8Encoder *self;
-    if ((self = (Ppmd8Encoder*)type->tp_alloc(type, 0)) != NULL) {
-        assert(self->inited == 0);
-        /* Thread lock */
-        if ((self->lock = PyThread_allocate_lock()) != NULL) {
-            return (PyObject*)self;
-        }
-        Py_XDECREF(self);
+    self = (Ppmd8Encoder*)type->tp_alloc(type, 0);
+     if (self == NULL) {
+        return NULL;
+     }
+    assert(self->inited == 0);
+    /* Thread lock */
+    if ((self->lock = PyThread_allocate_lock()) == NULL) {
+        goto error;
     }
-    PyErr_NoMemory();
-    return NULL;
+    return (PyObject*)self;
+
+error:
+    Py_XDECREF(self);
+    return PyErr_NoMemory();
 }
 
 static void
@@ -1601,17 +1702,23 @@ PyDoc_STRVAR(Ppmd8Encoder_doc, "A PPMd compression algorithm.\n\n"
                                  "mem_size:  max memory size in bytes the compressor is able to use, bigger values improve compression,\n"
                                  "           raging from 10kB to physical memory size.\n"
                                  "           Default size is 16MB.\n"
+                                 "end_mark:  Boolean when using endmark for compression.\n"
+                                 "           Default is False.\n"
+                                 "           When True,  encoder use \x01 as escape character and \x00 as endmark.\n"
+                                 "           When source data has a byte \x01, it is converted to \x01\x01 first then compressed."
                                  );
 
 static int
 Ppmd8Encoder_init(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
 {
-    static char *kwlist[] = {"max_order", "mem_size", NULL};
+    static char *kwlist[] = {"max_order", "mem_size", "end_mark", NULL};
     PyObject *max_order = Py_None;
     PyObject *mem_size = Py_None;
+    Bool endmark = False;
+
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
-                                     "OO:Ppmd8Encoder.__init__", kwlist,
-                                     &max_order, &mem_size)) {
+                                     "OO|p:Ppmd8Encoder.__init__", kwlist,
+                                     &max_order, &mem_size, &endmark)) {
         goto error;
     }
 
@@ -1622,6 +1729,8 @@ Ppmd8Encoder_init(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
     }
     self->inited = 1;
 
+    self->endmark = endmark;
+
     unsigned long maximum_order = 6;
     unsigned long memory_size = 16 << 20;
 
@@ -1657,6 +1766,7 @@ Ppmd8Encoder_init(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
             goto success;
         }
         PyMem_Free(self->cPpmd8);
+        PyErr_NoMemory();
     }
 error:
     return -1;
@@ -1694,17 +1804,36 @@ Ppmd8Encoder_encode(Ppmd8Encoder *self,  PyObject *args, PyObject *kwargs) {
     writer.outBuffer = &out;
     self->cPpmd8->Stream.Out = (IByteOut *) &writer;
 
+    Bool result = True;
+    Py_BEGIN_ALLOW_THREADS
     for (UInt32 i = 0; i < data.len; i++){
+        // when endmark mode, escape 0x01.
+        if (self->endmark && (*((Byte *)data.buf + i) == 0x01)) {
+            Ppmd8_EncodeSymbol(self->cPpmd8, 0x01);
+            if (out.size == out.pos) {
+                if (OutputBuffer_Grow(&buffer, &out) < 0) {
+                    PyErr_SetString(PyExc_ValueError, "No memory.");
+                    result = False;
+                    break;
+                } else {
+                    writer.outBuffer = &out;
+                }
+            }
+        }
         Ppmd8_EncodeSymbol(self->cPpmd8, *((Byte *)data.buf + i));
         if (out.size == out.pos) {
             if (OutputBuffer_Grow(&buffer, &out) < 0) {
                 PyErr_SetString(PyExc_ValueError, "No memory.");
-                goto error;
+                result = False;
+                break;
             } else {
                 writer.outBuffer = &out;
             }
         }
     }
+    Py_END_ALLOW_THREADS
+    if (!result)
+        goto error;
 
     ret = OutputBuffer_Finish(&buffer, &out);
     RELEASE_LOCK(self);
@@ -1738,7 +1867,11 @@ Ppmd8Encoder_flush(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
     writer.Write = Write;
     writer.outBuffer = &out;
     self->cPpmd8->Stream.Out = (IByteOut *) &writer;
-    Ppmd8_EncodeSymbol(self->cPpmd8, -1);  // endmark
+    if (self->endmark) {
+        Ppmd8_EncodeSymbol(self->cPpmd8, 0x01);  // endmark sequence
+        Ppmd8_EncodeSymbol(self->cPpmd8, 0x00);
+    }
+    Ppmd8_EncodeSymbol(self->cPpmd8, -1);  // endmark for encoder
     Ppmd8_RangeEnc_FlushData(self->cPpmd8);
     ret = OutputBuffer_Finish(&buffer, &out);
 
