@@ -1,17 +1,15 @@
 import sys
 
 import psutil
-import pytest
 from hypothesis import given
 from hypothesis import strategies as st
 
 import pyppmd
 
 vmem = psutil.virtual_memory()
-MAX_SIZE = min(0xFFFFFFFF - 12 * 3, sys.maxsize, vmem.available)
+MAX_SIZE = min(0xFFFFFFFF - 12 * 3, sys.maxsize, vmem.available >> 2)
 
 
-@pytest.mark.skipif(sys.platform.startswith("win"), reason="hypothesis test on windows fails with unknown reason.")
 @given(
     obj=st.binary(min_size=1),
     max_order=st.integers(min_value=2, max_value=64),
@@ -28,7 +26,6 @@ def test_ppmd7_fuzzer(obj, max_order, mem_size):
     assert result == obj
 
 
-@pytest.mark.skipif(sys.platform.startswith("win"), reason="hypothesis test on windows fails with unknown reason.")
 @given(
     obj=st.binary(min_size=1),
     max_order=st.integers(min_value=2, max_value=64),
